/*************************************************
* Lowest Level MPI Algorithms Header File        *
* (C) 1999-2006 The Botan Project                *
*************************************************/

#ifndef BOTAN_MP_ASM_H__
#define BOTAN_MP_ASM_H__

#include <botan/mp_types.h>

#if (BOTAN_MP_WORD_BITS != 32)
   #error The mp_ia32 module requires that BOTAN_MP_WORD_BITS == 32
#endif

namespace Botan {

extern "C" {

/*************************************************
* Word Multiply                                  *
*************************************************/
inline word word_mul(word a, word b, word* carry)
   {
   asm("mull %1\n\t"        // a(in eax) * b(wherever) -> edx:eax
       "addl %5,%0\n\t"     // add c to low word (eax)
       "adcl $0,%2"         // add carry from previous to high word (edx)
       : "=a"(a), "=rm"(b), "=&d"(*carry)
       : "0"(a), "1"(b), "g"(*carry) : "cc");
   return a;
   }

/*************************************************
* Word Multiply/Add                              *
*************************************************/
inline void word_madd(word a, word b, word c, word d,
                      word* out_low, word* out_high)
   {
   asm("mull %1\n\t"        // a(in eax) * b(wherever) -> edx:eax
       "addl %5,%0\n\t"     // add c to low word (eax)
       "adcl $0,%2\n\t"     // add carry from previous to high word (edx)
       "addl %6,%0\n\t"     // add d to low word (eax)
       "adcl $0,%2"         // add carry from previous to high word (edx)
       : "=a"(*out_low), "=rm"(b), "=&d"(*out_high)
       : "0"(a), "1"(b), "g"(c), "g"(d) : "cc");
   }

/*************************************************
* Multiply-Add Accumulator                       *
*************************************************/
inline void word3_muladd(word* w2, word* w1, word* w0, word x, word y)
   {
   u64bit z = (u64bit)x * y + (*w0);
   *w0 = (word)z;

   word t1 = (word)(z >> BOTAN_MP_WORD_BITS);
   *w1 += t1;
   *w2 += (*w1 < t1) ? 1 : 0;
   }

/*************************************************
* Multiply-Add Accumulator                       *
*************************************************/
inline void word3_muladd_2(word* w2, word* w1, word* w0, word x, word y)
   {
   u64bit z = (u64bit)x * y;
   word t0 = (word)z;
   word t1 = (word)(z >> BOTAN_MP_WORD_BITS);

   *w0 += t0;
   *w1 += t1 + ((*w0 < t0) ? 1 : 0);
   *w2 += (*w1 < t1) ? 1 : 0;

   *w0 += t0;
   *w1 += t1 + ((*w0 < t0) ? 1 : 0);
   *w2 += (*w1 < t1) ? 1 : 0;
   }

}

}

#endif
